from typing import reveal_type

from caqtus.types.units import Quantity, ureg, Unit
import pint


if __name__ == "__main__":
    q = Quantity(1, "m")
    print(type(q.magnitude))
    units = Quantity(1, "m").units
    print(type(units))
    print(type(1 * units))



from caqtus.extension import Experiment
from caqtus.session.sql import PostgreSQLConfig
from caqtus.session.sql._session_maker import SQLiteConfig


if __name__ == "__main__":
    experiment_1 = Experiment()
    experiment_1.setup_default_extensions()
    experiment_1.configure_storage(PostgreSQLConfig.from_file("config.yaml"))
    experiment_1.launch_condetrol()



from __future__ import annotations

import importlib.resources
from collections.abc import Sequence
from typing import overload, Self

import pint._typing
import pint.facets
import pint.facets.nonmultiplicative.objects
import pint.facets.numpy.quantity
import pint.facets.numpy.unit
from pint.facets.plain.quantity import ScalarT

from caqtus.types.recoverable_exceptions import InvalidValueError

UnitLike = pint._typing.UnitLike


class Quantity[M: pint._typing.Magnitude](
    pint.facets.system.objects.SystemQuantity[M],
    pint.facets.numpy.quantity.NumpyQuantity[M],
    pint.facets.nonmultiplicative.objects.NonMultiplicativeQuantity[M],
    pint.facets.plain.PlainQuantity[M],
):

    @overload
    def __new__(
        cls, value: int | float, units: UnitLike | None = None
    ) -> Quantity[float]: ...

    @overload
    def __new__(cls, value: str, units: UnitLike | None = None) -> Self: ...

    @overload
    def __new__(  # type: ignore[misc]
        cls, value: Sequence[ScalarT], units: UnitLike | None = None
    ) -> Self: ...

    @overload
    def __new__(cls, value: Self, units: UnitLike | None = None) -> Self: ...

    def __new__(cls, value, units=None):
        if isinstance(value, int):
            return super().__new__(  # pyright: ignore[reportCallIssue]
                cls,
                float(value),  # pyright: ignore[reportArgumentType]
                units,
            )
        return super().__new__(cls, value, units)

    @property
    def units(self) -> Unit:
        u = super().units
        assert isinstance(u, Unit)
        return u

    def to_base_units(self) -> Quantity[M]:
        result = super().to_base_units()
        assert isinstance(result, Quantity)
        return result


class Unit(
    pint.facets.SystemRegistry.Unit,
    pint.facets.numpy.unit.NumpyUnit,
    pint.facets.nonmultiplicative.objects.NonMultiplicativeUnit,
    pint.facets.plain.PlainUnit,
):
    pass


class UnitRegistry(pint.UnitRegistry):
    Quantity = Quantity  # pyright: ignore[reportAssignmentType]
    Unit = Unit  # pyright: ignore[reportAssignmentType]


units_definition_file = importlib.resources.files("caqtus.types.units").joinpath(
    "units_definition.txt"
)

ureg = UnitRegistry(
    str(units_definition_file),
    autoconvert_offset_to_baseunit=True,
    cache_folder=":auto:",
)
unit_registry = ureg
pint.set_application_registry(unit_registry)


UndefinedUnitError = pint.UndefinedUnitError

DimensionalityError = pint.DimensionalityError
dimensionless = Quantity(1, "").units

TIME_UNITS = {"s", "ms", "µs", "us", "ns"}

FREQUENCY_UNITS = {
    "Hz",
    "kHz",
    "MHz",
    "GHz",
    "THz",
}

POWER_UNITS = {
    "W",
    "mW",
    "dBm",
}

DIMENSIONLESS_UNITS = {"dB"}

CURRENT_UNITS = {"A", "mA"}

VOLTAGE_UNITS = {"V", "mV"}

DISTANCE_UNITS = {"m", "mm", "µm", "um", "nm"}

DEGREE_UNITS = {"deg", "rad"}

UNITS = (
    TIME_UNITS
    | FREQUENCY_UNITS
    | POWER_UNITS
    | DIMENSIONLESS_UNITS
    | CURRENT_UNITS
    | VOLTAGE_UNITS
    | DISTANCE_UNITS
    | DEGREE_UNITS
)


class InvalidDimensionalityError(InvalidValueError):
    """Raised when a value has an invalid dimensionality.

    This error is raised when a value has an invalid dimensionality and the user
    should fix it.
    """

    pass



"""Contains the copy function to transfer data between two sessions."""

from typing import assert_type

from caqtus.session import (
    PureSequencePath,
    ExperimentSession,
    PathNotFoundError,
    PathIsNotSequenceError,
    PathIsSequenceError,
    PathHasChildrenError,
    PathIsRootError,
    State,
    SequenceStateError,
    SequenceNotCrashedError,
)
from caqtus.session._shot_id import ShotId
from caqtus.utils._result import (
    Success,
    Failure,
    is_failure,
    is_failure_type,
    is_success,
)


def copy_sequence(
    source: PureSequencePath,
    source_session: ExperimentSession,
    destination: PureSequencePath,
    destination_session: ExperimentSession,
) -> (
    Success[None]
    | Failure[PathNotFoundError]
    | Failure[PathIsNotSequenceError]
    | Failure[PathIsSequenceError]
    | Failure[PathHasChildrenError]
    | Failure[SequenceStateError]
):
    """Copy the sequence from the source to the destination."""

    sequence_stats_result = source_session.sequences.get_stats(source)
    if is_failure(sequence_stats_result):
        return sequence_stats_result
    stats = sequence_stats_result.value
    state = stats.state
    if state not in {State.DRAFT, State.FINISHED, State.INTERRUPTED, State.CRASHED}:
        return Failure(SequenceStateError())
    creation_date = source_session.paths.get_path_creation_date(source)
    assert not is_failure_type(creation_date, PathNotFoundError)
    assert not is_failure_type(creation_date, PathIsRootError)
    iterations = source_session.sequences.get_iteration_configuration(source)
    time_lanes = source_session.sequences.get_time_lanes(source)

    creation_result = destination_session.sequences.create(
        destination, iterations, time_lanes
    )
    if is_failure(creation_result):
        return creation_result
    destination_session.paths.update_creation_date(destination, creation_date.value)

    if state == State.DRAFT:
        return Success(None)
    set_state(destination, State.PREPARING, destination_session)
    destination_session.sequences.set_global_parameters(
        destination, source_session.sequences.get_global_parameters(source)
    )
    set_state(destination, State.RUNNING, destination_session)

    for shot_index in range(stats.number_completed_shots):
        shot_parameters = source_session.sequences.get_shot_parameters(
            source, shot_index
        )
        shot_data = source_session.sequences.get_all_shot_data(source, shot_index)
        shot_start_time = source_session.sequences.get_shot_start_time(
            source, shot_index
        )
        shot_stop_time = source_session.sequences.get_shot_end_time(source, shot_index)
        destination_session.sequences.create_shot(
            ShotId(destination, shot_index),
            shot_parameters,
            shot_data,
            shot_start_time,
            shot_stop_time,
        ).unwrap()

    if state == State.FINISHED:
        set_state(destination, State.FINISHED, destination_session)

    if state == State.INTERRUPTED:
        set_state(destination, State.INTERRUPTED, destination_session)

    if state == State.CRASHED:
        set_state(destination, State.CRASHED, destination_session)
        exception_result = source_session.sequences.get_exception(source)
        assert not is_failure_type(exception_result, PathNotFoundError)
        assert not is_failure_type(exception_result, PathIsNotSequenceError)
        assert not is_failure_type(exception_result, SequenceNotCrashedError)
        exception = exception_result.value

        if exception:
            assert is_success(
                destination_session.sequences.set_exception(destination, exception)
            )
    destination_session.sequences.update_start_and_end_time(
        destination, stats.start_time, stats.stop_time
    )

    return Success(None)


def set_state(path: PureSequencePath, state: State, session: ExperimentSession) -> None:
    result = session.sequences.set_state(path, state)
    assert not is_failure_type(result, PathNotFoundError)
    assert not is_failure_type(result, PathIsNotSequenceError)
    assert_type(result, Success[None])



from collections.abc import Sequence

from typing_extensions import TypeIs, reveal_type


def is_sequence_of_int(x: Sequence) -> TypeIs[Sequence[int]]:
    return all(isinstance(v, int) for v in x)


def f(x: Sequence[int] | Sequence[str]) -> None:
    if is_sequence_of_int(x):
        reveal_type(x)
    else:
        reveal_type(x)



from typing import Any, Optional, overload, TypeAlias

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeIs

from ..recoverable_exceptions import InvalidTypeError
from ..units import Quantity, Unit, dimensionless, UnitLike

ScalarAnalogValue: TypeAlias = float | Quantity[float]
ArrayAnalogValue: TypeAlias = NDArray[np.floating] | Quantity[NDArray[np.floating]]
AnalogValue: TypeAlias = ScalarAnalogValue | ArrayAnalogValue


class NotAnalogValueError(InvalidTypeError):
    pass


class NotQuantityError(InvalidTypeError):
    pass


def is_scalar_analog_value(value: Any) -> TypeIs[ScalarAnalogValue]:
    """Returns True if the value is a scalar analog value, False otherwise."""

    if isinstance(value, float):
        return True

    if isinstance(value, Quantity):
        return isinstance(value.magnitude, float)

    return False


def is_array_analog_value(value: Any) -> TypeIs[ArrayAnalogValue]:
    """Returns True if the value is an array analog value, False otherwise."""

    if isinstance(value, np.ndarray):
        return issubclass(value.dtype.type, np.floating)

    if isinstance(value, Quantity):
        return isinstance(value.magnitude, np.ndarray) and issubclass(
            value.magnitude.dtype.type, np.floating
        )

    return False


def is_analog_value(value: Any) -> TypeIs[AnalogValue]:
    """Returns True if the value is an analog value, False otherwise."""

    return is_scalar_analog_value(value) or is_array_analog_value(value)


def is_quantity(value: Any) -> TypeIs[Quantity]:
    """Returns True if the value is a quantity, False otherwise."""

    return isinstance(value, Quantity)


def get_unit(value: AnalogValue) -> Optional[Unit]:
    """Returns the unit of the value if it has one, None otherwise."""

    if isinstance(value, Quantity):
        return value.units
    return None


@overload
def magnitude_in_unit[
    M: float | NDArray[np.floating]
](value: Quantity[M], unit: Optional[UnitLike]) -> M: ...


@overload
def magnitude_in_unit(value: float, unit: Optional[UnitLike]) -> float: ...


@overload
def magnitude_in_unit[
    A: NDArray[np.floating]
](value: A, unit: Optional[UnitLike]) -> A: ...


def magnitude_in_unit(value, unit):
    """Return the magnitude of a value in the given unit."""

    if is_quantity(value):
        if unit is None:
            return value.to(dimensionless).magnitude
        return value.to(unit).magnitude
    else:
        if unit is None:
            return value
        else:
            value = value * dimensionless
            return value.to(unit).magnitude


@overload
def add_unit(magnitude: float, unit: None) -> float: ...


@overload
def add_unit(magnitude: NDArray[np.floating], unit: None) -> NDArray[np.floating]: ...


@overload
def add_unit(magnitude: float, unit: Unit) -> Quantity[float]: ...


@overload
def add_unit(
    magnitude: NDArray[np.floating], unit: Unit
) -> Quantity[NDArray[np.floating]]: ...


def add_unit(magnitude, unit):
    """Add a unit to a magnitude."""

    if unit is None:
        return magnitude
    return Quantity(magnitude, unit)  # pyright: ignore[reportReturnType]


def are_units_compatible(unit1: Optional[Unit], unit2: Optional[Unit]) -> bool:
    """Return True if the two units are compatible, False otherwise."""

    if unit1 is None:
        return unit2 is None
    if unit2 is None:
        return unit1 is None

    return unit1.is_compatible_with(unit2)



from __future__ import annotations

from collections.abc import Set
from typing import Optional, Literal

import anyio
import attrs
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QIcon, QColor, QPalette, QKeySequence
from PySide6.QtWidgets import (
    QWidget,
    QToolBar,
    QStackedWidget,
    QLabel,
    QHBoxLayout,
    QApplication,
)

from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    ExperimentSession,
    AsyncExperimentSession,
    TracebackSummary,
)
from caqtus.session import (
    PathNotFoundError,
    SequenceNotEditableError,
    PathIsNotSequenceError,
    State,
)
from caqtus.types.iteration import (
    IterationConfiguration,
    StepsConfiguration,
)
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from .sequence_widget_ui import Ui_SequenceWidget
from .._icons import get_icon
from .._parameter_tables_editor import ParameterNamespaceEditor
from .._sequence_iteration_editors import StepsIterationEditor
from ..timelanes_editor import TimeLanesEditor
from ..timelanes_editor.extension import CondetrolLaneExtensionProtocol
from ..._common.exception_tree import ExceptionDialog


class SequenceWidget(QWidget, Ui_SequenceWidget):
    """Widget for editing sequence parameters, iterations and time lanes.

    This widget is a tab widget with three tabs: one for defining initial parameters,
    one for editing how the parameters should be iterated over for the sequence, and
    one for editing the time lanes that specify how a given shot should be executed.

    This widget is (optionally) associated with a sequence and displays the iteration
    configuration and time lanes for that sequence.
    If the widget is not associated with a sequence, it will hide itself.

    When associated with a sequence, the widget is constantly watching the state of the
    sequence.
    If the sequence is not in the draft state, the iteration editor and time lanes
    editor
     will become read-only.
    If the sequence is in the draft state, the iteration editor and time lanes editor
    will become editable and any change will be saved.
    """

    sequence_start_requested = Signal(PureSequencePath)
    """A signal emitted when the start sequence button is clicked."""

    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        extension: CondetrolLaneExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        """Initializes the sequence widget.

        The sequence widget will initially be associated with no sequence.

        Args:
            session_maker: It is used to connect to the storage system in which to look
            for the sequence.
            extension: The extension that provides time lanes customization.
            parent: The parent widget.
        """

        super().__init__(parent)
        self.setupUi(self)
        self.session_maker = session_maker

        with self.session_maker() as session:
            device_configurations = dict(session.default_device_configurations)

        self._state: _State = _SequenceNotSetState()
        self._state_lock = anyio.Lock()

        self.parameters_editor = ParameterNamespaceEditor(self)
        self.parameters_editor.set_read_only(True)
        self.time_lanes_editor = TimeLanesEditor(
            extension,
            device_configurations,
            self,
        )
        self.iteration_editor = StepsIterationEditor(self)

        self.tabWidget.clear()
        self.tabWidget.addTab(self.parameters_editor, "&Globals")
        self.tabWidget.addTab(self.iteration_editor, "&Parameters")
        self.tabWidget.addTab(self.time_lanes_editor, "Time &lanes")

        self.setup_connections()

        self.tool_bar = QToolBar(self)
        self.status_widget = IconLabel(icon_position="left")

        self.warning_action = self.tool_bar.addAction(
            get_icon("mdi6.alert", color=QColor(205, 22, 17)), "warning"
        )
        self.warning_action.triggered.connect(self._on_warning_action_triggered)
        self.warning_action.setToolTip("Error")

        self.tool_bar.addWidget(self.status_widget)
        self.start_sequence_action = self.tool_bar.addAction(
            get_icon("start", color=Qt.GlobalColor.darkGreen), "start"
        )
        self.start_sequence_action.triggered.connect(self._on_start_sequence_requested)
        self.start_sequence_action.setShortcut(QKeySequence("F5"))
        self.interrupt_sequence_action = self.tool_bar.addAction(
            get_icon("stop"), "interrupt"
        )
        self.interrupt_sequence_action.setVisible(False)
        self.verticalLayout.insertWidget(0, self.tool_bar)
        self.stacked = QStackedWidget(self)
        self.stacked.addWidget(QWidget(self))
        self.stacked.addWidget(self.iteration_editor.toolbar)
        self.stacked.addWidget(self.time_lanes_editor.toolbar)
        self.stacked.setCurrentIndex(0)
        self.tool_bar.addSeparator()
        self.tool_bar.addWidget(self.stacked)

        self.tabWidget.currentChanged.connect(self.stacked.setCurrentIndex)

        self._transition(_SequenceNotSetState())

        self._exception_dialog = ExceptionDialog(self)

    def set_available_parameter_names(
        self, parameter_names: Set[DottedVariableName]
    ) -> None:
        """Set the names of the parameters that are defined externally."""

        self.iteration_editor.set_available_parameter_names(parameter_names)

    async def exec_async(self) -> None:
        while True:
            await self.watch_sequence()
            await anyio.sleep(10e-3)

    async def watch_sequence(self) -> None:
        if isinstance(self._state, _SequenceNotSetState):
            return
        elif isinstance(self._state, _SequenceSetState):
            old_state = self._state
            new_state = await _query_state_async(
                self._state.sequence_path, self.session_maker
            )
            # It could be that the widget state changed while we were fetching infos,
            # for example if set_sequence is called or the editors emitted editions.
            if old_state != self._state:
                return

            if new_state != self._state:
                self._transition(new_state)
        else:
            raise AssertionError("Invalid state")

    def _transition(self, new_state: _State) -> None:
        match new_state:
            case _SequenceNotSetState():
                self.setVisible(False)
                self.start_sequence_action.setEnabled(False)
                self.interrupt_sequence_action.setEnabled(False)
                self.warning_action.setVisible(False)
            case _SequenceSetState(
                iterations=iterations,
                time_lanes=time_lanes,
                sequence_state=state,
                sequence_path=path,
            ):
                if not isinstance(iterations, StepsConfiguration):
                    raise NotImplementedError(f"Only supports {StepsConfiguration}")
                self.iteration_editor.set_iteration(iterations)
                self.time_lanes_editor.set_time_lanes(time_lanes)
                self._set_status_widget(path, state)
                if isinstance(new_state, _SequenceEditableState):
                    self.start_sequence_action.setEnabled(True)
                    self.interrupt_sequence_action.setEnabled(False)
                    self.time_lanes_editor.set_read_only(False)
                    self.iteration_editor.set_read_only(False)
                    self.tabWidget.setTabEnabled(0, False)
                    self.warning_action.setVisible(False)
                elif isinstance(new_state, _SequenceNotEditableState):
                    self.start_sequence_action.setEnabled(False)
                    self.interrupt_sequence_action.setEnabled(False)
                    self.time_lanes_editor.set_read_only(True)
                    self.iteration_editor.set_read_only(True)
                    self.parameters_editor.set_parameters(new_state.parameters)
                    self.tabWidget.setTabEnabled(0, True)
                    if isinstance(new_state, _SequenceCrashedState):
                        self.warning_action.setVisible(True)
                    else:
                        self.warning_action.setVisible(False)

                self.setVisible(True)
        self._state = new_state

    def _on_warning_action_triggered(self) -> None:
        """Display the sequence traceback in a dialog."""

        assert isinstance(self._state, _SequenceCrashedState)

        color = QApplication.palette().color(QPalette.ColorRole.Accent).name()
        self._exception_dialog.set_message(
            f"An error occurred while running the sequence "
            f"<b><font color='{color}'>{self._state.sequence_path}</font></b>."
        )
        traceback = self._state.exception_traceback
        self._exception_dialog.set_exception(traceback)
        self._exception_dialog.show()

    def _set_status_widget(self, path: PureSequencePath, state: State) -> None:
        text = " > ".join(path.parts)
        color = self.palette().text().color()
        if state.is_editable():
            icon = get_icon("editable-sequence", color=color)
        else:
            icon = get_icon("read-only-sequence", color=color)
        self.status_widget.set_text(text)
        self.status_widget.set_icon(icon)

    def _on_start_sequence_requested(self):
        assert isinstance(self._state, _SequenceEditableState)
        self.sequence_start_requested.emit(self._state.sequence_path)

    def set_sequence(self, sequence_path: Optional[PureSequencePath]) -> None:
        if sequence_path is None:
            new_state = _SequenceNotSetState()
        else:
            new_state = _query_state_sync(sequence_path, self.session_maker)
        self._transition(new_state)

    def setup_connections(self):
        self.time_lanes_editor.time_lanes_edited.connect(self.on_time_lanes_edited)
        self.iteration_editor.iteration_edited.connect(
            self.on_sequence_iteration_edited
        )

    def on_sequence_iteration_edited(self, iterations: IterationConfiguration):
        assert isinstance(self._state, _SequenceEditableState)
        with self.session_maker() as session:
            try:
                session.sequences.set_iteration_configuration(
                    self._state.sequence_path, iterations
                )
            except (PathNotFoundError, PathIsNotSequenceError):
                self._transition(_SequenceNotSetState())
            except SequenceNotEditableError:
                self._transition(
                    _query_state_sync(self._state.sequence_path, self.session_maker)
                )
            else:
                self._state = attrs.evolve(self._state, iterations=iterations)

    def on_time_lanes_edited(self, time_lanes: TimeLanes):
        assert isinstance(self._state, _SequenceEditableState)
        with self.session_maker() as session:
            try:
                session.sequences.set_time_lanes(self._state.sequence_path, time_lanes)
            except (PathNotFoundError, PathIsNotSequenceError):
                self._transition(_SequenceNotSetState())
            except SequenceNotEditableError:
                self._transition(
                    _query_state_sync(self._state.sequence_path, self.session_maker)
                )
            else:
                self._state = attrs.evolve(self._state, time_lanes=time_lanes)


class _State:
    pass


@attrs.frozen
class _SequenceNotSetState(_State):
    pass


@attrs.frozen
class _SequenceSetState(_State):
    sequence_path: PureSequencePath
    sequence_state: State

    iterations: IterationConfiguration
    time_lanes: TimeLanes


@attrs.frozen
class _SequenceNotEditableState(_SequenceSetState):
    parameters: ParameterNamespace


@attrs.frozen
class _SequenceCrashedState(_SequenceNotEditableState):
    exception_traceback: Optional[TracebackSummary]


@attrs.frozen
class _SequenceEditableState(_SequenceSetState):
    pass


async def _query_state_async(
    path: PureSequencePath, session_maker: ExperimentSessionMaker
) -> _State:
    async with session_maker.async_session() as session:
        is_sequence_result = await session.sequences.is_sequence(path)
        try:
            is_sequence = is_sequence_result.unwrap()
        except PathNotFoundError:
            return _SequenceNotSetState()
        else:
            if is_sequence:
                return await _query_sequence_state_async(path, session)
            else:
                return _SequenceNotSetState()


async def _query_sequence_state_async(
    path: PureSequencePath, session: AsyncExperimentSession
) -> _SequenceSetState:
    # These results can be unwrapped safely because we checked that the sequence
    # exists in the session.
    state = (await session.sequences.get_state(path)).unwrap()
    iterations = await session.sequences.get_iteration_configuration(path)
    time_lanes = await session.sequences.get_time_lanes(path)

    if state.is_editable():
        return _SequenceEditableState(
            path, iterations=iterations, time_lanes=time_lanes, sequence_state=state
        )
    else:
        parameters = await session.sequences.get_global_parameters(path)
        if state == State.CRASHED:
            tb_summary = (await session.sequences.get_traceback_summary(path)).unwrap()
            return _SequenceCrashedState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
                exception_traceback=tb_summary,
            )
        else:
            return _SequenceNotEditableState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
            )


def _query_state_sync(
    path: PureSequencePath, session_maker: ExperimentSessionMaker
) -> _State:
    with session_maker() as session:
        is_sequence_result = session.sequences.is_sequence(path)
        try:
            is_sequence = is_sequence_result.unwrap()
        except PathNotFoundError:
            return _SequenceNotSetState()
        else:
            if is_sequence:
                return _query_sequence_state_sync(path, session)
            else:
                return _SequenceNotSetState()


def _query_sequence_state_sync(
    path: PureSequencePath, session: ExperimentSession
) -> _SequenceSetState:
    # These results can be unwrapped safely because we checked that the sequence
    # exists in the session.
    state = session.sequences.get_state(path).unwrap()
    iterations = session.sequences.get_iteration_configuration(path)
    time_lanes = session.sequences.get_time_lanes(path)

    if state.is_editable():
        return _SequenceEditableState(
            path, iterations=iterations, time_lanes=time_lanes, sequence_state=state
        )
    else:
        parameters = session.sequences.get_global_parameters(path)
        if state == State.CRASHED:
            traceback_summary = session.sequences.get_exception(path).unwrap()
            return _SequenceCrashedState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
                exception_traceback=traceback_summary,
            )
        else:
            return _SequenceNotEditableState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
            )


class IconLabel(QWidget):
    def __init__(
        self,
        parent: Optional[QWidget] = None,
        icon_position: Literal["left", "right"] = "left",
    ):
        super().__init__(parent)
        self._label = QLabel()
        self._icon = QLabel()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        if icon_position == "left":
            layout.addWidget(self._icon)
            layout.addWidget(self._label)
        else:
            layout.addWidget(self._label)
            layout.addWidget(self._icon)
        self.setLayout(layout)

    def set_text(self, text: str):
        self._label.setText(text)

    def set_icon(self, icon: Optional[QIcon]):
        if icon is None:
            self._icon.clear()
        else:
            self._icon.setPixmap(icon.pixmap(20, 20))



import sqlalchemy
from sqlalchemy import select

from caqtus.session.sql._path_table import SQLSequencePath


if __name__ == "__main__":
    top_query = (
        select(SQLSequencePath)
        .where(SQLSequencePath.parent_id == 5)
        .cte(recursive=True)
    )
    alias = sqlalchemy.orm.aliased(top_query)
    descendants_query = top_query.union(
        select(SQLSequencePath).join(alias, SQLSequencePath.parent_id == alias.c.id)
    )
    query = select(alias).select_from(descendants_query)

    print(query.compile())



"""Provides an implementation of experiment sessions using SQL databases."""

from ._session_maker import (
    PostgreSQLExperimentSessionMaker,
    PostgreSQLConfig,
    SQLiteConfig,
)

__all__ = ["PostgreSQLExperimentSessionMaker", "PostgreSQLConfig", "SQLiteConfig"]



from __future__ import annotations

from collections.abc import Callable
from typing import Never, Literal

import attrs
from typing_extensions import TypeIs


@attrs.frozen
class Success[T]:
    value: T

    def unwrap(self) -> T:
        return self.value

    def map[R](self, func: Callable[[T], R]) -> Success[R]:
        return Success(func(self.value))

    @staticmethod
    def is_success() -> Literal[True]:
        return True

    @staticmethod
    def is_failure() -> Literal[False]:
        return False


def is_success[T, E: Exception](result: Result[T, E]) -> TypeIs[Success[T]]:
    return result.is_success()


def is_failure[T, E: Exception](result: Result[T, E]) -> TypeIs[Failure[E]]:
    return result.is_failure()


def is_failure_type[
    E: Exception
](result: Result, error_type: type[E]) -> TypeIs[Failure[E]]:
    return is_failure(result) and isinstance(result.error, error_type)


@attrs.frozen
class Failure[E: Exception]:
    error: E

    def unwrap(self) -> Never:
        raise self.error

    def map(self, func: Callable) -> Failure[E]:
        return self

    @staticmethod
    def is_success() -> Literal[False]:
        return False

    @staticmethod
    def is_failure() -> Literal[True]:
        return True


type Result[T, E: Exception] = Success[T] | Failure[E]
